# $Id: Protocol.pm,v 1.23 2003/02/06 22:39:20 blackketter Exp $

# SLIMP3 Server Copyright (C) 2001 Sean Adams, Slim Devices Inc.
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
package SliMP3::Protocol;

use strict;

use IO::Socket;
use IO::Select;

use SliMP3::Misc;

# The following settings are for testing the new streaming protocol.
# They allow easy simulation of latency and packet loss in the receive direction.
# This is just to for *very* basic testing - for real network simulation, use Dummynet.
my $SIMULATE_RX_LOSS  = 0.00;  # packet loss, 0..1
my $SIMULATE_RX_DELAY = 000;  # delay, milliseconds

#-------- You probably don't want to change this -----------------#
my $OLDSERVERPORT = 1069;   # Port that 1.2 and earlier clients use
my $SERVERPORT = 3483;		# IANA-assigned port for the SliMP3 protocol, used in firmware 1.3+
#-----------------------------------------------------------------#

# Select objects and select array ref returns
use vars qw(
	    $selUDPRead
	    $oldudpsock
	    $udpsock);

# This is all there is to it - it's like tftp, except that the client requests each
# chunk, as opposed to acknowlegding the data as we send it. Client handles timeouts.
sub gotRequest {

# FIXME - it would be nice to have per-client and server-wide statistics on bytes sent,
# packets dropped, average data rate, etc.

	my ($client, @octets) = @_;

	my $wptrhp = $octets[2];		# see protocol docs
	my $wptrlp = $octets[3];

	my $wptrh  = unpack 'C',  $wptrhp;
	my $wptrl  = unpack 'C',  $wptrlp;

	my $wptrhh = unpack 'H2', $wptrhp;
	my $wptrlh = unpack 'H2', $wptrlp;

	my $wptr   = $wptrh * 256 + $wptrl;

#	$client->usage($wptrh / 255.0);

	$::d_protocol_verbose && msg("Request wptr = $wptr - $wptrhh.$wptrlh\n");

	if  ($client->waitforstart()) {	# if we're waiting for the client to request the first chunk
					# of a new stream:

		if ($wptr != 0 ) {
			# if $wptr!=0 then this is probably a stray packet from the last stream we were playing
			# just log it and do nothing (it's normal).

			$::d_protocol && msg("Ignoring request from previous stream\n");
			return;

		} else {
			$client->waitforstart(0);
		}
	}

	# if we're not currently playing anything, just ignore the client.
	# FIXME - shouldn't we send the stop command?
	if (SliMP3::Playlist::playmode($client) ne "play") {
		$::d_protocol && msg("got a request, but we're not playing anything.\n");
	}

	my $chunkRef = undef;
	
	# packet was dropped, and the client requested a resend.
	if ($client->prevwptr() == $wptr) {
		$::d_protocol && msg("Duplicate request: wptr = $wptr\n");
		$chunkRef = SliMP3::Playlist::lastChunk($client);
	} else {
		$chunkRef = SliMP3::Playlist::nextChunk($client, SliMP3::Prefs::get('udpChunkSize'));
	}

	if (!defined($chunkRef)) {
		return;
	};
	# pad to a length of 18
	my $header = sprintf('m %-16s', $wptrhp.$wptrlp);

	my $append = '';
	
	# We must send an even number of bytes.
	if ((length($$chunkRef) % 2) != 0) {
		$append = '.';
	}

	sendClient($client, $header . $$chunkRef . $append);

	$::d_protocol_verbose && msg($client->id . " " . Time::HiRes::time() . " Sending ".length($$chunkRef . $append)." bytes. wptr = $wptrhh.$wptrlh\n");

	$client->prevwptr($wptr);
}

sub sendClient {
	my $client = shift;
	if (defined $client->udpsock()) {
		return $client->udpsock()->send( shift, 0, $client->paddr());
	} else {
		if ($::d_protocol) {
			bt();
			die $client->id . " no udpsock ready for client";
		}
	}	
}

sub processMessage {
	my ($client,$msg) = @_;

	my @octets = split(//,$msg);
	my $type   = unpack('a',$octets[0]);

	# for V2.0 or newer, all packets have the client's MAC address in [12..17]
	# in V1.3 and earlier, this value is undefined.

	if (defined($client->revision) && ($client->revision >= 2.0)) {
		$client->MAC_Address(join(':', unpack('H2', $octets[12]),
						unpack('H2', $octets[13]),
						unpack('H2', $octets[14]),
						unpack('H2', $octets[15]),
						unpack('H2', $octets[16]),
						unpack('H2', $octets[17])
					));
		$::d_protocol_verbose && msg("Client MAC: ".$client->MAC_Address."\n");
	}


	#  Packet format for IR code. Numbers are unsigned, network order.
	#
	#  [0]       'i'
	#  [1]       0x00
	#  [2..5]    player's time since startup in 'ticks' @ 625 KHz
	#  [6]       0xFF  (will eventually be an identifier for different IR code sets)
	#  [7]       number of bits ( always 16 for JVC )
	#  [8..11]   the IR code, up to 32 bits
	#  [12..17]  reserved/ignored

	if ($type eq 'i') {
		

		if ($::d_ir) {
			msg('irCode: [');

			for (my $oi = 0; $oi <= 11; $oi++) {
				msg(unpack('H2',$octets[$oi]), ' ');
			}

			msg("]\n");
		}

		# extract the IR code and the timestamp for the IR message
		my $irCodeBytes;
		$irCodeBytes = unpack('H*', $octets[8].$octets[9].$octets[10].$octets[11]);

		my $irTime  = unpack('N',$octets[2].$octets[3].$octets[4].$octets[5]);
		
		SliMP3::IR::process($client, $irCodeBytes, $irTime);

	} elsif ($type eq 'r') {

		gotRequest($client, @octets);

		SliMP3::Playlist::checkSync($client);

	} elsif ($type eq 'h') {

		SliMP3::Discovery::gotHello($client,
					    unpack('C',$octets[1]),
					    unpack('C',$octets[2])
					   );

	} elsif ($type eq '2') {

		# NOP - just the client acknowledging and i2c command.

	} elsif ($type eq 'd') {

		SliMP3::Discovery::gotDiscoveryRequest($client);

	} elsif ($type eq 'a') {

		my $wptr = unpack('C', $octets[6]) * 256 + unpack('C', $octets[7]);
		my $rptr = unpack('C', $octets[8]) * 256 + unpack('C', $octets[9]);
		my $seq  = unpack('C', $octets[10]) * 256 + unpack('C', $octets[11]);

		SliMP3::Stream::gotAck($client, $wptr, $rptr, $seq);

		SliMP3::Playlist::checkSync($client);

	} else {
		$::d_protocol && msg("debug: unknown type: [$type]\n");
	}

	return 1;
}

sub init {
	$selUDPRead = IO::Select->new();

# FIXME: Add a setup option to bind the server to a particular IP:PORT
#
	$udpsock = IO::Socket::INET->new(
		Proto     => 'udp',
		LocalPort => $SERVERPORT,
		LocalAddr => $main::localClientNetAddr
	);
	
	if (!$udpsock) {
		msg("Problem: There is already another copy of the SliMP3 server running on this machine.\n");
		exit 1;
	}

	$selUDPRead->add($udpsock); #to allow full processing of all pending UDP requests
	$::selRead->add($udpsock);

	$oldudpsock = IO::Socket::INET->new(
		Proto     => 'udp',
		LocalPort => $OLDSERVERPORT,
		LocalAddr => $main::localClientNetAddr
	);

	if (!$oldudpsock) {
		warn "Unable to open UDP socket on port $OLDSERVERPORT.  Make sure all clients are upgraded to version 1.3 or above.";
	} else {
		$selUDPRead->add($oldudpsock); #to allow full processing of all pending UDP requests
		$::selRead->add($oldudpsock);
	}
	
# say hello to the old clients that we might remember...
	my $clients = SliMP3::Prefs::get("clients");
	if (defined($clients)) {
		foreach my $addr (split( /,/, $clients)) {
			#make sure any new preferences get set to default values
			SliMP3::Discovery::sayHello($udpsock, $addr);
			
			#throttle the broadcasts
			select(undef,undef,undef,0.05);
		}
	}
}

sub handleClients {

	my $clientpaddr;
	my $msg = '';
	my $sock;
	
	my $start = Time::HiRes::time();
	# handle UDP activity...
	while ($sock=pending()) { #process all pending UDP messages on both old and new UDP ports
		my $now = Time::HiRes::time();
		if (($now - $start) > 0.5) {
			$::d_perf && msg("stayed in handleClients too long...\n");
			last;
		}
		$clientpaddr = recv($sock,$msg,1500,0);

		# simulate random packet loss
		next if ($SIMULATE_RX_LOSS && (rand() < $SIMULATE_RX_LOSS));

		if ($clientpaddr) {
			# Playlist::executecommand can be accessed over the UDP port
			if ($msg=~/^executecommand\((.*)\)$/) {
				my $ecArgs=$1;
				my @ecArgs=split(/, ?/, $ecArgs);
				$::d_protocol && msg("UDP: executecommand($ecArgs)\n");
				my $clientipport = shift(@ecArgs);
				my $client = SliMP3::Client::getClient($clientipport);
				SliMP3::Playlist::executecommand($client, \@ecArgs);
			} else {
			# it's Slim Protocol, coming from a client
				# check that it's a message type we know: starts with i r 2 d a or h (but not h followed by 0x00 0x00)
				if ($msg =~ /^([ir2da]|h[^\x00][^\x00])/) {
					my $client = getClient($clientpaddr, $sock);
	
					if ($::d_protocol_verbose) {
						my ($clientport, $clientip) = sockaddr_in($clientpaddr);
						msg("Client: ".inet_ntoa($clientip).":$clientport\n");
					}
	
					if ($SIMULATE_RX_DELAY) {
						# simulate rx delay
						SliMP3::Timers::setTimer($client, $now + $SIMULATE_RX_DELAY/1000,
									\&SliMP3::Protocols::processMessage, $msg);
					} else {
						processMessage($client, $msg);
					}
	
					# if a client is sending us data, but we don't know his ID/revision, say hello
					if (!$client->revision) {
						$::d_protocol && msg("reveived a packet from an unidentified client. saying hello.\n");
						SliMP3::Discovery::sayHello($sock, $client->id);
					}
				} else {
					if ($::d_protocol) {
						my ($clientport, $clientip) = sockaddr_in($clientpaddr);
						msg("ignoring Client: ".inet_ntoa($clientip).":$clientport that sent bogus message $msg\n");
					}
				}
			}
		} else {
			last;
		}
	}
}

###################
# return the client based on IP address and socket.  will create a new one if
# necessary 
sub getClient {
	my ($clientpaddr,$udpsock) = @_;

	# return the cached copy if it exists
	my ($port, $ip) = sockaddr_in($clientpaddr);
	$ip = inet_ntoa($ip);
	my $address = $ip.":".$port;
	
	my $client = SliMP3::Client::getClient($address);
	
	if (!defined($client)) {
		$client = SliMP3::Client::newClient($address);
		$client->paddr($clientpaddr);
		$client->udpsock($udpsock);
		$client->deviceid('slimp3');

		# add the new client all the currently known clients so we can say hello to them later
		my $clientlist = SliMP3::Prefs::get("clients");
	
		if (defined($clientlist)) {
			$clientlist .= ",$address";
		} else {
			$clientlist = $address;
		}
	
		my %seen = ();
		my @uniq = ();
		foreach my $item (split( /,/, $clientlist)) {
			push(@uniq, $item) unless $seen{$item}++ || $item eq '';
		}
		SliMP3::Prefs::set("clients", join(',', @uniq));
		
		# fire it up!
		SliMP3::Client::power($client,SliMP3::Prefs::clientGet($client, 'power'));

		# start the screen saver
		SliMP3::Buttons::ScreenSaver::screenSaver($client);
	}
	
	return $client
}

sub pending {
	my @readablesocks = ($selUDPRead->can_read(0));
	return(shift(@readablesocks));
}
