#!/usr/bin/perl -w

# SLIMP3 Server Copyright (C) 2001 Sean Adams, Slim Devices Inc.
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#

require 5.006_000;
use strict;  # _NOPERL2EXE_
#use utf8;

# This package section is used for the windows service version of the application, 
# as built with ActiveState's PerlSvc

package PerlSvc;

our $Name = 'slimp3svc';
our $DisplayName = 'SLIMP3 Server';

sub Startup {
	main::init();
	
    # here's where your startup code will go
    while (ContinueRun() && !main::idle()) {
    }
}

sub Install {
    # add your additional install messages or functions here
}

sub Remove {
    # add your additional remove messages or functions here
}

sub Help {
    # add your additional help messages or functions here
    $DisplayName . $Name;
}

package main;

use FindBin qw($Bin);
use File::Spec::Functions qw(:ALL);

use lib (catdir($Bin,'lib'), $Bin, catdir($Bin,'lib','CPAN'));

use Config;
use Getopt::Long;
use FileHandle;
use POSIX qw(:signal_h :errno_h :sys_wait_h);
use Socket qw(:DEFAULT :crlf);

use SliMP3::AlarmClock;
use SliMP3::Animation;
use SliMP3::Background;
use SliMP3::Buttons::Browse;
use SliMP3::Buttons::BrowseMenu;
use SliMP3::Buttons::Home;
use SliMP3::Buttons::Power;
use SliMP3::Buttons::ScreenSaver;
use SliMP3::Client;
use SliMP3::Control;
use SliMP3::CLI;
use SliMP3::Discovery;
use SliMP3::Display;
use SliMP3::HTTP;
use SliMP3::IR;
use SliMP3::Info;
use SliMP3::iTunes;
use SliMP3::MusicFolderScan;
use SliMP3::Misc;
use SliMP3::OSDetect;
use SliMP3::Playlist;
use SliMP3::Prefs;
use SliMP3::Protocol;
use SliMP3::Scheduler;
use SliMP3::Setup;
use SliMP3::Stdio;
use SliMP3::Strings qw(string);
use SliMP3::Timers;

use vars qw($VERSION
		@AUTHORS);

@AUTHORS = (

	'Sean Adams',
	'Dean Blackketter',
	'Amos Hayes',
	'Mark Langston',
	'Eric Lyons',
	'Scott McIntyre',
	'Robert Moser',
	'Richard Smith',
	'Sam Saffron',
	'Daniel Sully',
);

$VERSION = '3.1b2';

# old preferences settings, only used by the .slimp3.conf configuration.
# real settings are stored in the new preferences file:  .slimp3.pref
use vars qw(
	    $mp3dir
	    $httpport
);

use vars qw(
		$d_cli
		$d_control
		$d_display
		$d_files
		$d_http
		$d_info
		$d_ir
		$d_itunes
		$d_mdns
		$d_os
		$d_perf
		$d_parse
		$d_paths
		$d_playlist
		$d_protocol
		$d_prefs
		$d_remotestream
		$d_scan
		$d_server
		$d_scheduler
		$d_stdio
		$d_stream
		$d_stream_v
		$d_sync
		$d_time
		$d_ui
		$d_usage

	    $cliaddr
	    $cliport
	    $daemon
	    $httpaddr
	    $lastlooptime
	    $loopcount
	    $loopsecond
	    $localClientNetAddr
	    $localStreamAddr
	    $pidfile
	    $prefsfile
	    $quiet
	    $nosetup
	    $stdio
	);

# these IO::Select objects contain references to every socket we might be listening
# or connected to, so we can wait in select for network events
use vars qw(
	    $selRead
		$selWrite
	);
	

sub init {
	srand();

	$selRead = IO::Select->new();
	$selWrite = IO::Select->new();

	autoflush STDERR;
	autoflush STDOUT;

	SliMP3::OSDetect::init();

	SliMP3::Strings::init(catdir($Bin,'strings.txt'), "EN");

	if (SliMP3::OSDetect::OS() ne 'mac') {
		$SIG{CHLD} = 'IGNORE';
		$SIG{PIPE} = 'IGNORE';
		if (SliMP3::OSDetect::OS() ne 'win') {
			$SIG{INT} = \&sigint;
		}		
		$SIG{TERM} = \&sigint;
		$SIG{QUIT} = \&sigint;
	}

	initSettings();

	SliMP3::Strings::setLanguage(SliMP3::Prefs::get("language"));
	SliMP3::IR::init();
	SliMP3::Buttons::Common::init();

	$::d_server && msg("SLIMP3 Server starting up...\n");

	# background if requested
	if ($daemon) {
		daemonize();
	} else {
		save_pid_file();
	};
	
	if ($stdio) {
		SliMP3::Stdio::init(\*STDIN, \*STDOUT);
	}

	SliMP3::Protocol::init();
	SliMP3::Info::init();
	SliMP3::HTTP::init();
	SliMP3::CLI::init();
	
# start scanning based on a timer...
# Currently, it's set to one item (directory or song) scanned per second.
	if (SliMP3::iTunes::useiTunesLibrary()) {
		SliMP3::iTunes::startScan();
	} else {
		SliMP3::MusicFolderScan::startScan(1);
	}
	
	$lastlooptime = Time::HiRes::time();
	$loopcount = 0;
	$loopsecond = int($lastlooptime);
}

sub main {
	initOptions();

	init();
	
	while (!idle()) {}
	
	stopServer();
}


sub idle {

	my $stopFlag = 0;

	my $select_time;

	my $now = Time::HiRes::time();

	if (int($now) == $loopsecond) {
		$loopcount++;
	} else {
		$::d_perf && msg("Idle loop speed: $loopcount iterations per second\n");
		$loopcount = 0;
		$loopsecond = int($now);
	}
	
	my $to;
	if ($::d_perf) { $to = watchDog(); }
	
	# check for time travel (i.e. If time skips backwards for DST or clock drift adjustments)
	if ($now < $lastlooptime) {
		SliMP3::Timers::adjustAllTimers($now - $lastlooptime);
		$::d_time && msg("finished adjustalltimers: " . Time::HiRes::time() . "\n");
	} 
	$lastlooptime = $now;

	# check the timers for any new tasks		
	SliMP3::Timers::checkTimers();	
	if ($::d_perf) { $to = watchDog($to, "checkTimers"); }

	# handle SliMP3 client protocol activity
	SliMP3::Protocol::handleClients();
	if ($::d_perf) { $to = watchDog($to, "handleClients"); }
	
	# check the timers for any new tasks		
	SliMP3::Timers::checkTimers();	
	if ($::d_perf) { $to = watchDog($to, "checkTimers"); }

	my $tasks = SliMP3::Scheduler::run_tasks();
	if ($::d_perf) { $to = watchDog($to, "run_tasks"); } 
	
	# if background tasks are running, don't wait in select.
	if (!$tasks) {
		# undefined if there are no timers, 0 if overdue, otherwise delta to next timer
		$select_time = SliMP3::Timers::nextTimer();

		$::d_time && msg("select_time: ". (defined($select_time) ? $select_time : "UNDEF")."\n");
		
		# wait in select until we get something.  
		$::d_time && msg("select counts: " . $selRead->count . " " . $selWrite->count . "\n");

		my ($r, $w, $e) = IO::Select->select($selRead,$selWrite,undef,$select_time);
		if ($::d_perf) { $to = watchDog($to, "select"); }
		
		if ($stdio) {
			foreach my $fh (@$r) {
				if ($fh == $SliMP3::Stdio::stdin) {
				  SliMP3::Stdio::processRequest($fh);
				}
			}
		}
	}
	
	# handle HTTP and command line interface activity, including:
	#   opening sockets, 
	#   reopening sockets if the port has changed, 
	#   and handling HTTP traffic
	SliMP3::HTTP::idle();
	if ($::d_perf) { $to = watchDog($to, "http::idle"); }

	SliMP3::CLI::idle();
	if ($::d_perf) { $to = watchDog($to, "cli::idle"); }

	return $stopFlag;
}

sub showUsage {
	print <<EOF;
Usage: $0 [--mp3dir <dir>] [--daemon] [--stdio]
          [--httpport <portnumber> [--httpaddr <listenip>]]
          [--cliport <portnumber> [--cliaddr <listenip>]]
          [--prefsfile <prefsfilepath> [--pidfile <pidfilepath>]]
          [--d_various]

    --help           => Show this usage information.
    --mp3dir         => The path to a directory of your MP3 files.
    --daemon         => Run the server in the background.
                        This may only work on Unix-like systems.
    --stdio          => Use standard in and out as a command line interface 
                        to the server
    --httpport       => Activate the web interface on the specified port.
                        Set to 0 in order disable the web server.
    --httpaddr       => Activate the web interface on the specified IP address.
    --cliport        => Activate the command line interface TCP/IP interface
                        on the specified port. Set to 0 in order disable the 
                        command line interface server.
    --cliaddr        => Activate the command line interface TCP/IP 
                        interface on the specified IP address.
    --prefsfile      => Specify where the preferences file should be stored
    --pidfile        => Specify where a process ID file should be stored
    --quiet          => Minimize the amount of text output
    --playeraddr     => Specify the _server's_ IP address to use to connect 
                        to SLIMP3 players
    --streamaddr     => Specify the _server's_ IP address to use to connect
                        to streaming audio sources
    --nosetup        => Disable setup via http.

The following are debugging flags which will print various information 
to the console via stderr:

    --d_control      => Low level player control information
    --d_cli          => Display debugging information for the 
                        command line interface interface
    --d_display      => Show what (should be) on the slimp3's display 
    --d_files        => Files, paths, opening and closing
    --d_http         => HTTP activity
    --d_info         => MP3/ID3 track information
    --d_ir           => Infrared activity
    --d_itunes       => iTunes synchronization information
    --d_mdns         => Multicast DNS aka Zeroconf aka Rendezvous information
    --d_os           => Operating system detection information
    --d_paths        => File path processing information
    --d_perf         => Performance information
    --d_parse        => Playlist parsing information
    --d_playlist     => High level playlist and control information
    --d_protocol     => Client protocol information
    --d_prefs        => Preferences file information
    --d_remotestream => Information about remote HTTP streams and playlists
    --d_scan         => Information about scanning directories and filelists
    --d_server       => Basic server functionality
    --d_scheduler    => Internal scheduler information
    --d_stdio        => Standard I/O command debugging
    --d_stream       => Information about player streaming protocol 
    --d_stream_v     => Verbose information about player streaming protocol 
    --d_sync         => Information about multi player synchronization
    --d_time         => Internal timer information
    --d_ui           => Player user interface information
    --d_usage        => Display buffer usage codes on the SLIMP3 display
    
Commands may be sent to the server through standard in and will be echoed via
standard out.  See complete documentation for details on the command syntax.
EOF

}

sub initOptions {
	if (!GetOptions(
		'cliaddr=s'   		=> \$cliaddr,
		'cliport=s'   		=> \$cliport,
		'daemon'   			=> \$daemon,
		'httpaddr=s'   		=> \$httpaddr,
		'httpport=s'   		=> \$httpport,
		'mp3dir=s' 			=> \$mp3dir,
		'pidfile=s' 		=> \$pidfile,
		'playeraddr=s'		=> \$localClientNetAddr,
		'stdio'				=> \$stdio,
		'streamaddr=s'		=> \$localStreamAddr,
		'prefsfile=s' 		=> \$prefsfile,
		'quiet'   			=> \$quiet,
		'nosetup'			=> \$nosetup,
		'd_cli'				=> \$d_cli,
		'd_control'			=> \$d_control,
		'd_display'			=> \$d_display,
		'd_files'			=> \$d_files,
		'd_http'			=> \$d_http,
		'd_info'			=> \$d_info,
		'd_ir'				=> \$d_ir,
		'd_itunes'			=> \$d_itunes,
		'd_mdns'			=> \$d_mdns,
		'd_os'				=> \$d_os,
		'd_paths'			=> \$d_paths,
		'd_perf'			=> \$d_perf,
		'd_parse'			=> \$d_parse,
		'd_playlist'		=> \$d_playlist,
		'd_protocol'		=> \$d_protocol,
		'd_prefs'			=> \$d_prefs,
		'd_remotestream'	=> \$d_remotestream,
		'd_scan'			=> \$d_scan,
		'd_server'			=> \$d_server,
		'd_scheduler'		=> \$d_scheduler,
		'd_stdio'			=> \$d_stdio,
		'd_stream'			=> \$d_stream,
		'd_stream_v'		=> \$d_stream_v,
		'd_sync'			=> \$d_sync,
		'd_time'			=> \$d_time,
		'd_ui'				=> \$d_ui,
		'd_usage'			=> \$d_usage,
	)) {
		showUsage();
		exit(1);
	};
}

sub initSettings {	
	SliMP3::Prefs::load($prefsfile);
	SliMP3::Prefs::checkServerPrefs();
	SliMP3::Buttons::Home::updateMenu();
	SliMP3::Setup::initSetup();
	
	#options override existing preferences
	if (defined($mp3dir)) {
		SliMP3::Prefs::set("mp3dir", $mp3dir);
	}
	
	if (defined($httpport)) {
		SliMP3::Prefs::set("httpport", $httpport);
	}

	if (defined($cliport)) {
		SliMP3::Prefs::set("cliport", $cliport);
	}

	# warn if there's no mp3dir preference
	# FIXME put the strings in strings.txt
	if (!(defined SliMP3::Prefs::get("mp3dir") && 
				-d SliMP3::Prefs::get("mp3dir")) && 
				!$quiet && 
				!SliMP3::iTunes::useiTunesLibrary()) {
		msg("Your MP3 directory needs to be configured. Please open your web browser,\n");
		msg("go to the following URL, and click on the \"Server Settings\" link.\n\n");
		msg(string('SETUP_URL_WILL_BE') . "\n\t" . SliMP3::HTTP::HomeURL() . "\n");
	} else {
		if (SliMP3::Prefs::get("mp3dir") =~ m|[/\\]$|) {
			$mp3dir = SliMP3::Prefs::get("mp3dir");
			$mp3dir =~ s|[/\\]$||;
			SliMP3::Prefs::set("mp3dir",$mp3dir);
		}
	}
}

sub daemonize {
	my $pid; 
	use POSIX 'setsid';
	
	if (!open STDIN, '/dev/null') { die "Can't read /dev/null: $!";}
	if (!open STDOUT, '>/dev/null') { die "Can't write to /dev/null: $!";}
	if (!defined($pid = fork)) { die "Can't fork: $!"; }
	
	if ($pid) {
		save_pid_file($pid);
		# don't clean up the pidfile!
		$pidfile = undef;
		exit;
	}
	
	$0 = "slimp3d";
	
	if (!setsid) { die "Can't start a new session: $!"; }
	if (!open STDERR, '>&STDOUT') { die "Can't dup stdout: $!"; }
 }

#------------------------------------------
#
# Clean up resources and exit.
#
sub stopServer {
	$::d_server && msg("SLIMP3 server shutting down.\n");
	cleanup();
	exit();
}

sub sigint {
	cleanup();
	exit();
}

sub cleanup {
	if (SliMP3::Prefs::get('usetagdatabase')) {
		SliMP3::Info::stopCache();
	}
	remove_pid_file();
	SliMP3::mDNS::stopAdvertise();
}

sub save_pid_file {
	 my $process_id = shift || $$;


	 if (defined $pidfile && -e $pidfile) {
	 	die "Process ID file: $pidfile already exists";
	 }
	 
	 if (defined $pidfile and open PIDFILE, ">$pidfile") {
		print PIDFILE "$process_id\n";
		close PIDFILE;
	 }
}
 
sub remove_pid_file {
	 if (defined $pidfile) {
	 	unlink $pidfile;
	 }
}
 
sub END {
	sigint();
}

# start up the server if we're not running as a service.	
if (!defined($PerlSvc::VERSION)) { 
	main()
};



__END__
